# perl
=comm
| -----------------------------------------------------------------------
| Script: Fracture.pl
| Author: Dmitry Bersenev aka Mynglam
| Version 0.8  (4 apr 2014)
| Last changes:
| 2 feb 2012:	"Abort" now working
| 14 mar 2012:	rewriting slice system
|				add Making UV and Material
|				add Align centers
| 2 apr 2014:   fix fracture (crack) problem in 701
| 4 apr 2014:   UV makes in selected Vertex Map
| -----------------------------------------------------------------------
|     (  ).
|           .
| -----------------------------------------------------------------------
| Cut into pieces mesh (must be closed).
| As the centers of the chunks may be used vertices not belonging to polygons.
| -----------------------------------------------------------------------
=cut

lxout("Fractured START!");

our $app_version = lxq("query platformservice appversion ?");

my $chunks = lxq( "user.value llama_fracturechunks ?");
my $uservertex = lxq( "user.value llama_fracture_uservertex ?");
my $seed = lxq( "user.value llama_fractureseed ?");
our $makeslicemat = lxq( "user.value llama_fracture_makemat ?");
our $makesliceuv = lxq( "user.value llama_fracture_makeuv ?");
our $aligncenters = lxq( "user.value llama_fracture_aligncenters ?");

our $materialpref = "Slice";
our $partforslice = "Slice_tmp";
our $slice_part = "Slice_part";
our $boolerrors = 0;
our $boolmessage = "";

srand($seed);

my @lrs = lxq("query layerservice layers ? fg");
my $userlayer = lxq("query layerservice layers ? main");
if(@lrs ==0 || !$userlayer) { die("No active layers"); }
my $userlayerid = lxq("query layerservice layer.id ? $userlayer");
my $originname = lxq("query layerservice layer.name ? $userlayer");

#     
my @ed = lxq("query layerservice edges ? all");
if(@ed == 0) { die; }
foreach(@ed)
{
	if (lxq("query layerservice edge.numPolys ? $_") != 2)
	{ die("Mesh must be closed"); }
}

#lx("select.typeFrom item");
lx("select.subItem $userlayerid set mesh 0 0");
lx("item.duplicate type:locator");
#my $itembox = GetItem();
#my $box_id = lxq("query sceneservice item.id ? $itembox");
#lx("select.subItem $box_id set mesh 0 0");
lx("transform.freeze");

my $originlayer = lxq("query layerservice layers ? main");
our $originid = lxq("query layerservice layer.id ? $originlayer");

my @bounds = lxq("query layerservice layer.bounds ? $originid");
my @X = ($bounds[0], $bounds[3]);
my @Y = ($bounds[1], $bounds[4]);
my @Z = ($bounds[2], $bounds[5]);
my $sizebox = Max($bounds[3]-$bounds[0], $bounds[4]-$bounds[1], $bounds[5]-$bounds[2]);
lxout("Bounding Box Size = $sizebox");
lxout("BBox X = @X");
lxout("BBox Y = @Y");
lxout("BBox Z = @Z");

local %cores;
local %sortcores;

if ($uservertex)
{
	my @verts = lxq("query layerservice verts ? all");
	$chunks = 0;
	foreach(@verts)
	{
		my @num = lxq("query layerservice vert.polyList ? $_");
		if(@num == 0)
		{
			$chunks++;
			my @pos = lxq("query layerservice vert.pos ? $_");
			push @{$cores{$chunks}}, @pos;
		}
	}
	if($chunks < 2) { die("Vertices must be at least two."); }
}
else
{
	for(1..$chunks)
	{
		my @pos;
		push @pos, rand($X[1]-$X[0]) + $X[0];
		push @pos, rand($Y[1]-$Y[0]) + $Y[0];
		push @pos, rand($Z[1]-$Z[0]) + $Z[0];
		push @{$cores{$_}}, @pos;
	}
}

PreserveToolsAttr();

lxout("Chunks: $chunks");
lxmonInit($chunks);

foreach $key (keys %cores)
{
	my %sort;
	#         
	foreach $skey (keys %cores)
	{
		if($skey == $key) { next; }
		$sort{$skey} = Length(@{$cores{$key}}, @{$cores{$skey}});
		#lxout("[$skey]: $sort{$skey}");
	}
	#     ( )
	#    
	my @sorted = sort{$sort{$a} <=> $sort{$b}} keys %sort;
	push @{$sortcores{$key}}, @sorted;
	
	#lxout("Core[$key]: (@{$sortcores{$key}})");
}

#      
my $namelocator = "$originname Fracture";
lx("item.create groupLocator");
my $item = GetItem();
our $groupid = lxq("query sceneservice item.id ? $item");
#lx("item.parent $groupid $originid inPlace:1");
lx("item.name \"$namelocator\" groupLocator $groupid");

for(1..$chunks)
{
	local $sli = 0;
	#lxout("Chunk $_ slicing...");
	#       
	lx("select.subItem $originid set mesh 0 0");
	lx("item.duplicate type:locator");
	$item = GetItem();
	#      
	#my $id = lxq("query sceneservice item.id ? $item");
	local $id = lxq("query sceneservice item.id ? $item");
	lx("item.parent $id $groupid inPlace:1");
	lx("item.name Chunk_$_ mesh $id");
	local $chunk_name = lxq("query sceneservice item.name ? $id");
	
	#      ,   item
	if (Sliced($_) == 0)
	{
		lx("select.subItem $id set mesh 0 0");
		lx("item.delete");
		lxout("Chunk_$_ empty -> delete this.");
	}
	lxout("Chunk_$_ sliced $sli time.");
	if(!lxmonStep()) { die("User Abort"); }
}

DivideChunks();
AlignCenters();

lx("select.subItem $originid set mesh 0 0");
lx("item.delete");
lx("select.subItem $userlayerid set mesh 0 0");
lx("workPlane.reset");

RestoreToolsAttr();
lxout("Fracture FINISH!");

if($boolerrors > 0)
{
	lx("dialog.setup warning");
	lx("dialog.title {Fracture Errors}");
	lx("dialog.msg {Some boolean operations were errors.\nTry to simplify the mesh or change \"Seed\".\nMeshes, which had errors($boolerrors):\n$boolmessage}");
	lx("dialog.open");
}

sub Length
{
	my @vec = ($_[3]-$_[0], $_[4]-$_[1], $_[5]-$_[2]);
	return sqrt($vec[0]*$vec[0]+$vec[1]*$vec[1]+$vec[2]*$vec[2]);
}

sub Sliced # 0 - chunk
{
	my $chunk = shift;
	my @sort = @{$sortcores{$chunk}};
	#lxout("Chunk $chunk: @sort");
	
	#       
	lx("select.typeFrom vertex");
	lx("select.drop vertex");
	lx("select.vertex add poly equal 0");
	lxq("query layerservice layers ? main");
	my @vrt = lxq("query layerservice verts ? selected");
	if(@vrt != 0) { lx("delete"); }
	
	foreach $ind (@sort)
	{
		$sli++;
		
		if (Slice($chunk, @{$cores{$ind}}))
		{
			last;
		}
	
		#        item
		#lx("select.typeFrom polygon");
	}
	
	lxq("query layerservice layers ? main");
	
	if (!lxq("query layerservice poly.N ? all")) {
		return 0; }
		
	return 1;
}

sub Slice
{
	my $chunk = shift;
	my @corepos;
	push @corepos, shift, shift, shift;
	my @chunkpos = @{$cores{$chunk}};
	
	lx("workPlane.reset");
	
	my @planepos = (($corepos[0]+$chunkpos[0])/2, ($corepos[1]+$chunkpos[1])/2, ($corepos[2]+$chunkpos[2])/2);
	my @normal = ($corepos[0]-$planepos[0], $corepos[1]-$planepos[1], $corepos[2]-$planepos[2]);
	
	#          
	if ($sli > 10)
	{
		my $layer = lxq("query layerservice layers ? main");
		my @bound = lxq("query layerservice layer.bounds ? $layer");
		my $x = Max(abs($bound[0] - $chunkpos[0]), abs($bound[3] - $chunkpos[0]));
		my $y = Max(abs($bound[1] - $chunkpos[1]), abs($bound[4] - $chunkpos[1]));
		my $z = Max(abs($bound[2] - $chunkpos[2]), abs($bound[5] - $chunkpos[2]));
		if (Length(@normal) > Length($x, $y, $z))
		{
			return 1;
		}
	}
	
	lx("select.typeFrom polygon");
	lx("select.drop polygon");
	lx("select.typeFrom vertex");
	lx("select.drop vertex");
	
	my @vvec = (0,0,0);
	my @vvec2 = (0,0,0);
	
	#       
	if($normal[2] != 0)
	{
		$vvec[0] = rand();
		$vvec[1] = rand();
		$vvec[2] = ($normal[0]*$vvec[0] + $normal[1]*$vvec[1])/(-$normal[2]);
		$vvec2[0] = rand();
		$vvec2[1] = rand();
		$vvec2[2] = ($normal[0]*$vvec2[0] + $normal[1]*$vvec2[1])/(-$normal[2]);
	}
	elsif($normal[1] != 0)
	{
		$vvec[0] = rand();
		$vvec[2] = rand();
		$vvec[1] = ($normal[0]*$vvec[0] + $normal[2]*$vvec[2])/(-$normal[1]);
		$vvec2[0] = rand();
		$vvec2[2] = rand();
		$vvec2[1] = ($normal[0]*$vvec2[0] + $normal[2]*$vvec2[2])/(-$normal[1]);
	}
	elsif($normal[0] != 0)
	{
		$vvec[1] = rand();
		$vvec[2] = rand();
		$vvec[0] = ($normal[1]*$vvec[1] + $normal[2]*$vvec[2])/(-$normal[0]);
		$vvec2[1] = rand();
		$vvec2[2] = rand();
		$vvec2[0] = ($normal[1]*$vvec2[1] + $normal[2]*$vvec2[2])/(-$normal[0]);
	}
	else
	{
		lxout("Shit happen.");
		return;
	}
	@vvec = Normalize(@vvec);
	@vvec2 = Normalize(@vvec2);
	#lxout("XYZ @vvec / @vvec2");
	
	my @vvert = ($vvec[0]+$planepos[0], $vvec[1]+$planepos[1], $vvec[2]+$planepos[2]);
	my @vvert2 = ($vvec2[0]+$planepos[0], $vvec2[1]+$planepos[1], $vvec2[2]+$planepos[2]);

	#     
	lx("tool.set prim.makeVertex on 0");
	lx("tool.setAttr prim.makeVertex cenX $planepos[0]");
	lx("tool.setAttr prim.makeVertex cenY $planepos[1]");
	lx("tool.setAttr prim.makeVertex cenZ $planepos[2]");
	lx("tool.doApply");
	lx("select.nextMode");
	lx("tool.set prim.makeVertex flush 0");
		
	lx("tool.set prim.makeVertex on 0");
	lx("tool.setAttr prim.makeVertex cenX $vvert[0]");
	lx("tool.setAttr prim.makeVertex cenY $vvert[1]");
	lx("tool.setAttr prim.makeVertex cenZ $vvert[2]");
	lx("tool.doApply");
	lx("select.nextMode");
	lx("tool.set prim.makeVertex flush 0");
	
	lx("tool.set prim.makeVertex on 0");
	lx("tool.setAttr prim.makeVertex cenX $vvert2[0]");
	lx("tool.setAttr prim.makeVertex cenY $vvert2[1]");
	lx("tool.setAttr prim.makeVertex cenZ $vvert2[2]");
	lx("tool.doApply");
	lx("select.nextMode");
	lx("tool.set prim.makeVertex flush 0");
	
	lx("poly.make auto");
	lx("select.typeFrom polygon");
	
	if (NormalOrient(Normalize(@normal)) == -1) { lx("poly.flip"); }
	
	lx("workPlane.fitSelect");
	lx("delete");
	lx("select.drop polygon");

	#     
	my $maxbox = $sizebox*4;
	lx("layer.new");
	my $bool_layer = GetItem();
	my $bool_id = lxq("query sceneservice item.id ? $bool_layer");
	lx("tool.set prim.cube on");
	lx("tool.reset");
	lx("tool.attr prim.cube minX -$maxbox");
	lx("tool.attr prim.cube minY 0.0");
	lx("tool.attr prim.cube minZ -$maxbox");
	lx("tool.attr prim.cube maxX $maxbox");
	lx("tool.attr prim.cube maxY $maxbox");
	lx("tool.attr prim.cube maxZ $maxbox");
	lx("tool.attr prim.cube segmentsX 1");
	lx("tool.attr prim.cube segmentsY 1");
	lx("tool.attr prim.cube segmentsZ 1");
	lx("tool.attr prim.cube radius 0.0");
	lx("tool.attr prim.cube uvs false");
	lx("tool.apply");
	lx("tool.set prim.cube off");
	lx("poly.setPart \"$partforslice\"");
	
	lx("select.subItem $id set mesh 0 0");
	lx("select.all");
	lx("select.subItem $bool_id add mesh 0 0");
	my $boolcheck = lx("!poly.boolean subtract lastselect");
	lx("select.drop polygon");
	lx("!vert.merge fixed false 0.001 false true");
	lx("select.subItem $bool_id set mesh 0 0");
	lx("item.delete");
	lx("select.subItem $id set mesh 0 0");

	MeshCleanup();
	FindHoles();
	my $lr = lxq("query layerservice layers ? main");
	my $pall = lxq("query layerservice poly.N ? all");
	lx("select.typeFrom polygon");
	lx("select.drop polygon");
	if($pall !=0)
	{
		lx("!select.polygon add part face $partforslice");
		my $psel = lxq("query layerservice poly.N ? selected");
		if($psel != 0 && $pall-$psel > 0)
		{
			MakeUV();
		}
	}
	lx("workPlane.reset");
	
	if(!$boolcheck) {
		$boolerrors++;
		$boolmessage = "$boolmessage\n$chunk_name";
		return 1; }
	return 0;
}

sub Max
{
	my $max = $_[0];
	foreach(@_)
	{
		if ($_ > $max) { $max = $_; }
	}
	return $max;
}

#       ()
sub NormalOrient
{
	my @cn;
	push @cn, shift, shift, shift;
	my $lay = lxq("query layerservice layers ? main");
	my @polys = lxq("query layerservice polys ? selected");
	my @pnorm = lxq("query layerservice poly.normal ? $polys[0]");
	#my $err = 0;
	@pnorm = Normalize(@pnorm);
	$cn[0] = $pnorm[0] + $cn[0];
	$cn[1] = $pnorm[1] + $cn[1];
	$cn[2] = $pnorm[2] + $cn[2];
	#if(abs($pnorm[0]-$cn[0]) > 0.01) { $err++; }
	#if(abs($pnorm[1]-$cn[1]) > 0.01) { $err++; }
	#if(abs($pnorm[2]-$cn[2]) > 0.01) { $err++; }
	#if($err != 0) { return -1; }
	if(Length(@cn) <  0.5) { return -1; }
	else { return 1; }
}

sub MakeUV
{
	#   
	my @vmaps = lxq("query layerservice vmaps ? texture");
	if(@vmaps)
	{       
			#    ,   ,     
			my $vmapsId = -1;
			my $i = @vmaps;
			while($vmapsId < 0 and $i > 0)
			{
				$i--;
				if (lxq("query layerservice vmap.selected ? @vmaps[$i]"))
				{
					$vmapsId = $i;
				}
			
			}
			if($vmapsId < 0) {$vmapsId = $vmaps[0]}
			my $vmap_name = lxq("query layerservice vmap.name ? $vmapsId");
			lx("!select.vertexMap $vmap_name txuv replace");
			#$vmap = lxq("query layerservice vmaps ? selected");
	}
	
	lx("select.typeFrom polygon");
	lx("select.drop polygon");
	lx("!select.polygon add part face $partforslice");
	
	if($makeslicemat) { lx("poly.setPart Default"); }
	else { lx("poly.setPart $slice_part"); }
	
	#   
	if($makeslicemat)
	{ lx("poly.setMaterial \"$materialpref $originname\" {1.0 1.0 1.0} 0.8 0.2 false false"); }
	
	#    UV 
	if($makesliceuv)
	{	
		lx("tool.set uv.create on");
		lx("tool.reset");
		lx("tool.attr uv.create proj planar");
		lx("tool.attr uv.create axis 1");
		lx("tool.attr uv.create mode manual");
		lx("tool.attr uv.create sizX $sizebox");
		lx("tool.attr uv.create sizY $sizebox");
		lx("tool.attr uv.create sizZ $sizebox");
		lx("tool.attr uv.create cenX 0.0");
		lx("tool.attr uv.create cenY 0.0");
		lx("tool.attr uv.create cenZ 0.0");
		lx("tool.attr uv.create rotP 0.0");
		lx("tool.attr uv.create rotH 0.0");
		lx("tool.attr uv.create rotB 0.0");
		lx("tool.doApply");
		lx("tool.set uv.create off");
	}
}

sub Normalize
{
	my $len = sqrt($_[0]*$_[0]+$_[1]*$_[1]+$_[2]*$_[2]);
	if($len == 0) { return (0,0,0); }
	return ($_[0]/$len, $_[1]/$len, $_[2]/$len);
}

sub GetItem
{
	my $items = lxq("query sceneservice item.N ?");
	for(0..$items)
	{
		my $sel = lxq("query sceneservice item.isSelected ? $_");
		if($sel)
		{
			return $_;
		}
	}
	return undef;
}

sub MeshCleanup
{
	my $lr = lxq("query layerservice layers ? main");
	my $v_def = lxq("query layerservice vert.N ? all");
	if($v_def == 0) { return; }

	lx("select.typeFrom vertex");
	lx("select.drop vertex");
	lx("select.vertex add poly equal 0");
	lx("select.vertex add poly equal 1");
	lx("select.vertex add edge equal 1");
		
	my $v_sel = lxq("query layerservice vert.N ? selected");
	if($v_def > $v_sel && $v_sel > 0) {
		lx("remove");
		lxout("$v_sel vertex removed");
	}

	lx("select.typeFrom polygon");
	lx("select.drop polygon");
	lx("select.polygon add type colocated 8");
	$v_def = lxq("query layerservice vert.N ? all");
	$v_sel = lxq("query layerservice vert.N ? selected");
	if($v_def > $v_sel && $v_sel > 0) {
		lx("delete");
		lxout("$v_sel Co-Located polys deleted");
	}
}

sub FindHoles
{
	my $lay = lxq("query layerservice layers ? main");
	my $alledges = lxq("query layerservice edge.N ? all");
	if($alledges == 0) { return; }
	
	lx("select.typeFrom edge");
	lx("select.drop edge");
	lx("!select.edge add bond equal (none)");
	
	my $hedges = lxq("query layerservice edge.N ? selected");
	my @edges = lxq("query layerservice edges ? selected");
		
	if($alledges != $hedges && $hedges > 1)
	{
		lxout("Hole find (edges $hedges)");
		#my @v = lxq("query layerservice edge.vertList ? $edges[0]");
		#my @p = lxq("query layerservice edge.polyList ? $edges[0]");
			
		#lx("!select.element $lay edge set $v[0] $v[1] $p[0]");
		#lx("!select.loop");
		lx("!poly.make auto");
		lx("select.typeFrom polygon");
		#lx("poly.triple");
		lx("poly.setPart $partforslice");
		#lx("select.drop polygon");
		#lx("!poly.align");
	}
	else { return; }
}

sub DivideChunks
{
	my @kids = lxq("query sceneservice item.children ? $groupid");
	lxout("Kids: @kids");
	foreach(@kids)
	{
		while(1)
		{
			lx("select.subItem $_ set mesh 0 0");
			lx("select.typeFrom polygon");
			lx("select.drop polygon");
			my $layer = lxq("query layerservice layers ? main");
			lx("select.element $layer polygon set 0");
			lx("select.connect");
			my $allpolys = lxq("query layerservice poly.N ? all");
			my $selpolys = lxq("query layerservice poly.N ? selected");
			if($allpolys > $selpolys)
			{
				lx("cut");
				lx("layer.new");
				lx("paste");
				my $item = GetItem();
				my $id = lxq("query sceneservice item.id ? $item");
				lx("item.parent $id $groupid inPlace:1");
				lx("item.name Chunk mesh");
				lxout("Added another \"Chunk\"");
			}
			else { last; }
		}
	}
}

# align centers of chunks to center bbox
sub AlignCenters
{
	my @kids = lxq("query sceneservice item.children ? $groupid");
	#lxout("Kids: @kids");
	foreach(@kids)
	{
		if($aligncenters)
		{
			lx("select.subItem $_ set mesh 0 0");
			lx("select.typeFrom center");
			lx("select.center $_ set");
			lx("center.bbox center");
		}
		else
		{
			lx("select.subItem $_ set mesh 0 0");
			lx("item.match cen pos all 0 {$_} {$userlayerid}");
			lx("item.match cen rot all 0 {$_} {$userlayerid}");
			lx("item.match cen scl all 0 {$_} {$userlayerid}");
		}
	}
}

sub PreserveToolsAttr
{
	lx("tool.set prim.cube on");
	lx("tool.makePreset llama_cube");
	lx("tool.set prim.cube off");
	
	lx("tool.set uv.create on");
	lx("tool.makePreset llama_uv");
	lx("tool.set uv.create off");
}

sub RestoreToolsAttr
{
	lx("tool.set prim.cube on");
	lx("tool.restorePreset llama_cube");
	lx("tool.set prim.cube off");
	
	lx("tool.set uv.create on");
	lx("tool.restorePreset llama_uv");
	lx("tool.set uv.create off");
}

