# perl
=comm
| -----------------------------------------------------------------------
| Script: Locators2UV.pl
| Author: Dmitry Bersenev aka Mynglam
| Version 1.1  (14 march 2012)
| Last changes:
| 28 aug 2011:	Bake in different modes
| 14 mar 2012:	Preserve and restore tool attributes
| -----------------------------------------------------------------------
| Bake procedural texture locators in to the UV map.
| UV map is taken first on the list if there is none, it will be created.
| Projections type support: Cubic, Planar, Spherical, Cylindrical.
| -----------------------------------------------------------------------
|       UV .
| UV     ,   ,  .
|   : Cubic, Planar, Spherical, Cylindrical.
| -----------------------------------------------------------------------
=cut

my $PI = 3.1415926535897932384626433832795;

lxout("Baking type \"$ARGV[0]\"");

PreserveToolsAttr();

my @layers = lxq("query layerservice layers ? all");

local @selpolys = lxq("query layerservice polys ? selected");

my $items = lxq("query sceneservice item.N ?");
lxout("Items in scene $items");

foreach $layer (@layers)
{
	my $layrname = lxq("query layerservice layer.name ? $layer");
	my $visiblity = lxq("query layerservice layer.visible ? $layer");
	lxout("Layer _____________ $layrname [$visiblity]");
	
	if ($ARGV[0] eq "selected" && ($visiblity eq "bg" || $visiblity eq "none")) { next; }
	if ($ARGV[0] eq "active" && ($visiblity eq "none")) { next; }
	if ($ARGV[0] eq "main" && ($visiblity ne "main")) { next; }
		
	#    ,   
	foreach $item (0..$items-1)
	{
		my $name = lxq("query sceneservice item.name ? $item");
		my $type = lxq("query sceneservice item.type ? $item");
		#lxout("$name - $type");
		
		if ($type eq "txtrLocator")
		{
			my ($proj, $axis, @pos, @rot, @scale);			
			
			my @textures = lxq("query layerservice textures ? all");
			
			#    
			foreach $tex (@textures)
			{
				my $tex_name = lxq("query layerservice texture.name ? $tex");
				my $item_name = substr($name, 0, -10);
				
				#      
				if ($item_name eq $tex_name)
				{
					lxout("$item_name <=> $tex_name");
					
					#      
					$proj = lxq("query layerservice texture.projType ? $tex");
					$axis = lxq("query layerservice texture.projAxis ? $tex");
					@pos = lxq("query sceneservice item.pos ? $item");
					@rot = lxq("query sceneservice item.rot ? $item");
					@scale = lxq("query sceneservice item.scale ? $item");
					
					my $parent_ind;
					
					#        
					foreach $i (0..$items-1)
					{
						my $i_name = lxq("query sceneservice item.name ? $i");
						my $i_type = lxq("query sceneservice item.type ? $i");
						
						if ($i_type eq "imageMap" && $i_name eq $item_name)
						{
							$parent_ind = GetParent($i, "mask", 5);
							lxout("$parent_ind");
						}
					}
					
					if ($parent_ind == -1) { last; }
					my $parent = lxq("query sceneservice item.name ? $parent_ind");
					$parent = substr($parent, 0, -11);
					
					#   ,   UV
					Projection($layer, $parent, $proj, $axis, @pos, @rot, @scale);
					
					last;
				}
			}
		}
	}
}

RestoreToolsAttr();
lx("dialog.setup info");
lx("dialog.title {Locators to UV}");
lx("dialog.msg { Projection texture locators to UV is done! }");
lx("dialog.open");


#       ()
sub GetParent # 0 - index child, 1 - mask, 2 - cycle
{
	my $p_ind = lxq("query sceneservice item.parent ? $_[0]");
	my $p_type = lxq("query sceneservice item.type ? $p_ind");
	if ($_[1] eq $p_type) { return $p_ind; }
	if ($_[2] == 0) { return -1; }
	return GetParent($p_ind, $_[1], $_[2]-1);
}

#   UV     
sub Projection # 0 - layer, 1 - mat, 2 - proj, 3 - axis, 4-6 pos, 7-9, rot, 10-12 scale
{
	my $layer = shift;
	my $mat = shift;
	my $proj = shift;
	my $axis = shift;
	my (@pos, @rot, @scale);
	push @pos, shift, shift, shift;
	push @rot, shift, shift, shift;
	push @scale, shift, shift, shift;
	
	#   
	$rot[0] = ($rot[0] * 180) / $PI;
	$rot[1] = ($rot[1] * 180) / $PI;
	$rot[2] = ($rot[2] * 180) / $PI;
	
	if ($proj ne "cubic" && $proj ne "spherical" && $proj ne "planar" && $proj ne "cylindrical")
	{
		lxout("Working only {Cubic|Planar|Spherical|Cylindrical} projections");
		return;
	}
	
	lxout("$mat : $proj : $axis");
	#lxout("Position: @pos");
	#lxout("Rotation: @rot");
	#lxout("Scale: @scale");
	
	my @vmaps = lxq("query layerservice vmaps ? all");
	my $map_name = "none";
	
	#   UV 
	foreach $map (@vmaps)
	{
		my $map_type = lxq("query layerservice vmap.type ? $map");
		if ($map_type eq "texture")
		{
			$map_name = lxq("query layerservice vmap.name ? $map");
			#lxout("Select map: $map_name");
			last;
		}
	}
	
	my $layername = lxq("query layerservice layer.name ? $layer");
	
	lx("select.item {$layername} set");
	if($map_name ne "none") {	lx("select.vertexMap $map_name txuv replace")	};
	lx("select.typeFrom polygon");
	
	#    
	if($ARGV[0] eq "selected")
	{
		lxout("@selpolys");
		my @foruv;
		
		#        
		foreach $poly (@selpolys)
		{
			my $matp = lxq("query layerservice poly.material ? $poly");
			lxout("$matp");
			if($matp eq $mat)
			{
				push @foruv, $poly;
			}
		}
		
		#lxout("@foruv");
		lx("select.drop polygon");
		if(@foruv == 0)
		{
			#          
			return;
		}
		
		#     
		foreach (@foruv)
		{
			lx("select.element $layer polygon add $_");
		}				
	}
	else
	{
		lx("select.drop polygon");
		lx("select.polygon add material face $mat");
	}
		
	lx("tool.set uv.create on");
	lx("tool.reset");
	lx("!tool.attr uv.create axis $axis");
	lx("tool.attr uv.create mode manual");
	
	lx("tool.attr uv.create proj $proj");
	
	lx("tool.attr uv.create cenX $pos[0]");
	lx("tool.attr uv.create cenY $pos[1]");
	lx("tool.attr uv.create cenZ $pos[2]");
	
	lx("tool.attr uv.create rotP $rot[0]");
	lx("tool.attr uv.create rotH $rot[1]");
	lx("tool.attr uv.create rotB $rot[2]");
	
	lx("tool.attr uv.create sizX $scale[0]");
	lx("tool.attr uv.create sizY $scale[1]");
	lx("tool.attr uv.create sizZ $scale[2]");
	
	lx("tool.doApply");
	lx("tool.set uv.create off");
	#lx("select.nextMode");
}

sub PreserveToolsAttr
{
	lx("tool.set uv.create on");
	lx("tool.makePreset llama_uv");
	lx("tool.set uv.create off");
}

sub RestoreToolsAttr
{
	lx("tool.set uv.create on");
	lx("tool.restorePreset llama_uv");
	lx("tool.set uv.create off");
}





