# perl
=comm
| -----------------------------------------------------------------------
| Script: SmartTriangulation.pl
| Author: Dmitry Bersenev aka Mynglam
| Version 1.5 (11 march 2012)
| Last changes:
| 14 sep 2011:	fixed selection vertex map
| 2 feb 2012:	improve speed (hide geometry) 
| 				"Abort" now working
| 11 mar 2012:	add Angle Threshold
| -----------------------------------------------------------------------
| The script divides the selected polygons into triangles, or all if no selection.
| When it tries to keep the edges of the situation more adequately.
| Only works in the selection of polygons.
| -----------------------------------------------------------------------
|      ,      .
|        .
|      .
| -----------------------------------------------------------------------
=cut

local $PI = 3.1415926535897932384626433832795;

my $U_iterations = lxq( "user.value llama_iterations ?");
my $U_keepuv = lxq( "user.value llama_keepuvbounds ?");
my $U_keepmat = lxq( "user.value llama_keepmatbounds ?");
my $U_anglethreshold = lxq( "user.value llama_anglethreshold ?");
lxout("Iterations($U_iterations) | Mat($U_keepmat) | UV($U_keepuv)");

my $smode;

if (lxq("select.typeFrom typelist:{vertex;edge;polygon;item} ? ") )
{	$smode = "vertex";	}
elsif (lxq("select.typeFrom typelist:{edge;polygon;vertex;item} ? ") )
{	$smode = "edge";	} 
elsif (lxq("select.typeFrom typelist:{polygon;vertex;edge;item} ? ") )
{	$smode = "polygon";	}
else
{	$smode = "item";	}
  
lxout("Selection mode = $smode");

lx("select.typeFrom polygon");
lx("hide.unsel");
lx("poly.triple");

local $layer = lxq("query layerservice layers ? main");
#my $layer = lxq("query layerservice layers ? fg");

my $selall = 0;

local @userpolys = lxq("query layerservice polys ? selected");
if(@userpolys == 0)
{
	@userpolys = lxq("query layerservice polys ? all");
	if(@userpolys == 0) { die; }
	$selall = 1;
}
my $pc = scalar(@userpolys);
lxout("User polys = $pc");

lxmonInit($U_iterations*scalar(@userpolys));

for(1..$U_iterations)
{
	lxout("___ Iteration $_ ____");
	lx("select.typeFrom edge");
	lx("select.drop edge");
	#lx("select.edgeCoplanar 0.005");
	#lx("select.edgeCoplanar $U_anglethreshold");
	lx("select.edgeSharp 0.0 $U_anglethreshold");
	lx("select.edge remove bond equal (none)");

	my @staticedges = lxq("query layerservice edges ? unselected");
	my @liveedges = lxq("query layerservice edges ? selected");

	my @vmaps = lxq("query layerservice vmaps ? texture");
	if(@vmaps)
	{
		my $vmap_name = lxq("query layerservice vmap.name ? $vmaps[0]");
		lx("select.vertexMap $vmap_name txuv replace");
		$vmap = lxq("query layerservice vmaps ? selected");
	}
	#lxout("[$vmaps[0]] $vmap_name");

	#            
	foreach $ledge (@liveedges)
	{
		my @e_polys = lxq("query layerservice edge.polyList ? $ledge");
			
		if($U_keepuv)
		{
			my @p_discos = lxq("query layerservice poly.discos ? $e_polys[0]");
			push @p_discos, lxq("query layerservice poly.discos ? $e_polys[1]");
			my @e_vert = lxq("query layerservice edge.vertList ? $ledge");
			my $match = 0;
			
			foreach $ev (@e_vert)
			{
				foreach $pd (@p_discos)
				{
					if ($ev == $pd)
					{
						$match++;
						last;
					}
				}
			}
			
			if ($match >= 2)
			{
				push @staticedges, $ledge;
				#lxout("Edge-- $ledge");
			}
					
			#lxout("$ledge @e_polys | $match");
		}
		
		if($U_keepmat)
		{
			my $p_mt1 = lxq("query layerservice poly.material ? $e_polys[0]");
			my $p_mt2 = lxq("query layerservice poly.material ? $e_polys[1]");
			if ($p_mt1 ne $p_mt2)
			{
				push @staticedges, $ledge;
				#lxout("Edge-- $ledge");
			}
		}
	}

	#      ( ,      )
	lx("select.typeFrom edge");
	lx("select.drop edge");
	foreach $e (@staticedges)
	{
		my @v = lxq("query layerservice edge.vertList ? $e");
		my @p = lxq("query layerservice edge.polyList ? $e");
		lx("select.element $layer edge add $v[0] $v[1] $p[0]");
	}
	lx("select.editSet _temp add");
	lx("select.drop edge");
	lx("select.invert");
	lx("select.useSet _temp deselect");
	lx("select.deleteSet _temp");
	#   
	@liveedges = lxq("query layerservice edges ? selected");
	#lxout("Live edges: @liveedges");

	
	local @workpolys = @userpolys;
	local @dualpolys;
	local @neigh;
		
	while(1)
	{
		my $neighbour = FindNeighbour($workpolys[0], @liveedges);
		#lxout("$workpolys[0] -> $neighbour");
		if($neighbour)
		{
			push @neigh, $neighbour;
			push @dualpolys, $workpolys[0], $neighbour;
			#lxout("$workpolys[0] -> $neighbour");
		}
		else
		{
			$#neigh = -1;
			splice(@workpolys, 0, 1);
			if(!lxmonStep)
			{
				die("User Abort");
			}
		}
		if(@workpolys == 0) { last; }
	}
	
	#lxout("Dual: @dualpolys");
	Optimize(@dualpolys);
}

lx("select.typeFrom $smode");

sub Optimize
{
	my @polys = @_;
	
	for($i=0; $i<@polys-1; $i+=2)
	{
		lx("select.typeFrom polygon");
		lx("select.drop polygon");
		lx("select.element $layer polygon add $polys[$i]");
		lx("select.element $layer polygon add $polys[$i+1]");
		
		my @vertlist;
		my @p1verts = lxq("query layerservice poly.vertList ? $polys[$i]");
		my @p2verts = lxq("query layerservice poly.vertList ? $polys[$i+1]");
		
		my %data;
				
		foreach $pv1 (@p1verts)
		{
			foreach $pv2 (@p2verts)
			{
				$data{$pv1}++;
				$data{$pv2}++;
			}
		}
		
		if(keys(%data) > 4)
		{
			#lxout("Hash overhead");
			next;
		}
		
		while(my ($k,$v) = each(%data))
		{
			#lxout("Data: data[$k]=$v");
			if($v < 4) { push @vertlist, $k; }
		}
		
		while(my ($k,$v) = each(%data))
		{
			if($v > 4) { push @vertlist, $k; }
		}
		
		my $sum1 = SumAngle(@vertlist);
		@vertlist = ($vertlist[2], $vertlist[3], $vertlist[0], $vertlist[1]);
		my $sum2 = SumAngle(@vertlist);
		#lxout("[$polys[$i], $polys[$i+1]] $sum1 | $sum2");
		
		if($sum2 < $sum1)
		{
			lx("!!edge.spinQuads");
		}
	}
	return;
}

lx("select.typeFrom polygon");
lx("select.drop polygon");
if (!$selall)
{
	foreach $e (@userpolys)
	{
		lx("select.element $layer polygon add $e");
	}
}

lx("unhide");	

sub SumAngle
{
	my @v1 = lxq("query layerservice vert.pos ? $_[0]");
	my @v2 = lxq("query layerservice vert.pos ? $_[1]");
	my @vert1 = lxq("query layerservice vert.pos ? $_[2]");
	my @vert2 = lxq("query layerservice vert.pos ? $_[3]");
	
	my @vec1 = GetVector(@v1, @vert1);
	my @vec2 = GetVector(@v1, @vert2);
	my $angle1 = GetAngle(@vec1, @vec2);#*180/$PI;
	@vec1 = GetVector(@v2, @vert1);
	@vec2 = GetVector(@v2, @vert2);
	my $angle2 = GetAngle(@vec1, @vec2);#*180/$PI;
	return $angle1+$angle2;
}

sub GetAngle
{
	my @vec1 = (@_[0], @_[1], @_[2]);
	my @vec2 = (@_[3], @_[4], @_[5]);
	#lxout("V1: @vec1 | V2: @vec2");
	
	return acos(($vec1[0]*$vec2[0]+$vec1[1]*$vec2[1]+$vec1[2]*$vec2[2])/
				(sqrt($vec1[0]*$vec1[0]+$vec1[1]*$vec1[1]+$vec1[2]*$vec1[2])*
				sqrt($vec2[0]*$vec2[0]+$vec2[1]*$vec2[1]+$vec2[2]*$vec2[2])));
}

sub GetVector # 0, 1, 2, 3, 4, 5
{
	return ($_[3]-$_[0], $_[4]-$_[1], $_[5]-$_[2]);
}

sub acos { atan2( sqrt(1 - $_[0] * $_[0]), $_[0] ) } 

sub FindNeighbour # 0 - $poly, 1 - @liveedges
{
	my $poly = shift;
	my @verts = lxq("query layerservice poly.vertList ? $poly");
	#lxout("Poly $poly (@verts)");
	
	#     
	foreach $vert (@verts)
	{
		my @n_polys = lxq("query layerservice vert.polyList ? $vert");
		#lxout("Vert $vert (@n_polys)");
		
		#       
		foreach $npoly (@n_polys)
		{
			#         ,   
			if($npoly == $poly) {	next;	}
			#      - 
			if(CompareElement($npoly, @neigh)) { next; }
			
			#     (   )
			my $ct;
			foreach $wp (@userpolys)
			{
				if($wp == $npoly) { $ct++; }
			}
			if($ct == 0) {	next;	}
			
			my @nverts = lxq("query layerservice poly.vertList ? $npoly");
			my $count = 0;
			my @edge;
		
			foreach(@nverts)
			{
				if($_ == @verts[0] || $_ == @verts[1] || $_ == @verts[2])
				{
					push @edge, $_;
					$count++;
				}
				
				#          ,      
				if($count >= 2)
				{
					#lxout("Neighbour ($poly | $npoly) Edge(@edge)");
					#         
					foreach $live (@_)
					{
						my @liveedge = lxq("query layerservice edge.vertList ? $live");
						
						if (CompareEdge(@edge, @liveedge))
						{
							#lxout("Compare (@edge | @liveedge) - compare");
							return $npoly;
						}
					}
				}
			}
		}
	}
	
	return undef;
}

sub CompareElement
{
	my $val = shift;
	foreach(@_)
	{
		if($val == $_) { return 1; }
	}
	return 0;
}

sub CompareEdge
{
	if(	(@_[0] == @_[2] && @_[1] == @_[3]) ||
		(@_[0] == @_[3] && @_[1] == @_[2])	)
		{	return 1;	}
	return 0;
}



